unit u_rtfparser;

{<
  TrRTFParser is light version of RTF format parser.
  It simply extracts all tags and texts to the StringList which can be used
  to produce plain text file or use rtf tags to produce formatted text.

  Updated version compatible with Delphi2010+ and 64bit platform.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2017)
  @lastmod(Sep 2017)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, StrUtils;

type
  { TrRTFParser is light version of RTF format parser.
    It simply extracts all tags and texts to the StringList which can be used
    to produce plain text file or use rtf tags to produce formatted text. }
  TrRTFParser = class(TComponent)
  private
    FRtfContent: string;
    FRtfElements: TStrings;
    FFontTable: TStrings;
    TempTags: TStrings;
    Ch1, Ch2: Char;
    ResStr: string;
    IsTagActive: boolean;
    procedure AddChar;
    procedure AddResStr;
    function CheckTagStop: boolean;
    procedure ParseText(rtf: string);
    procedure ParseLines(SupportedTags: TStrings = nil);
    procedure ParseFontTbl(tag: string);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { return whether RtfContent contain rtf format }
    function IsRtfFormat: boolean;
    { parse RtfContent and fill RtfElements by supported tags and text;
      supported tags can use wild char * for any chars and # for any numbers }
    procedure ParseRtfContent(SupportedTags: TStrings = nil);
    { define RtfContent }
    property RtfContent: string read FRtfContent write FRtfContent;
    { contain result of parsing }
    property RtfElements: TStrings read FRtfElements;
    { contain list of defined fonts (f0, f1, f2) }
    property FontTable: TStrings read FFontTable;
  end;

implementation

uses rtool;

const
  cRTFDefTag = '{\rtf1';
  cRTFRtlch = '{\rtlch';
  cRTFFontTbl = '{\fonttbl';

constructor TrRTFParser.Create(AOwner: TComponent);
begin
  inherited;
  FRtfElements:=TStringList.Create;
  FFontTable:=TStringList.Create;
  TempTags:=TStringList.Create;
end;

destructor TrRTFParser.Destroy;
begin
  FRtfElements.Free;
  FFontTable.Free;
  TempTags.Free;
  inherited;
end;

procedure TrRTFParser.AddChar;
begin
  ResStr:=ResStr + Ch1;
end;

procedure TrRTFParser.AddResStr;
begin
  if ResStr <> '' then
    TempTags.Add(ResStr);
  ResStr:='';
end;

function TrRTFParser.CheckTagStop: boolean;
begin
  Result:=IsTagActive;
  if IsTagActive then
  begin
    AddResStr;
    IsTagActive:=false;
  end;
end;

function TrRTFParser.IsRtfFormat: boolean;
begin
  Result:=AnsiStartsText(cRTFDefTag, FRtfContent);
end;

procedure TrRTFParser.ParseRtfContent(SupportedTags: TStrings = nil);
begin
  FRtfElements.Clear;
  if not IsRtfFormat then
    Exit;
  ParseText(copy(FRtfContent, Length(cRTFDefTag) + 1, Length(FRtfContent)));
  ParseLines(SupportedTags);
end;

procedure TrRTFParser.ParseFontTbl(tag: string);
var
  p: integer;
  s, fnum, fname: string;
begin
  // remove starting tag
  tag:=rtool.FindStrPart(tag, cRTFFontTbl, 2);
  // remove spec tags (panose, fname)
  tag:=rtool.ReplaceStrFromTo('{\*\', '}', ' ', tag);
  repeat
    if tag = '' then
      Break;
    // remove one part between { and }
    rtool.RemoveStrPart(tag, '{\');
    s:=rtool.RemoveStrPart(tag, '}');
    // first part is font num id
    fnum:=Trim(rtool.FindStrPart(s, '\', 1));
    // find last tag + font name
    p:=LastDelimiter('\', s);
    s:=copy(s, p + 1, Length(s));
    // remove last tag, keep font name only
    rtool.RemoveStrPart(s, ' ');
    fname:=Trim(rtool.FindStrPart(s, ';', 1));
    // and name and value to fonttable
    if (fnum <> '') and (fname <> '') then
      FFontTable.Values[fnum]:=fname;
  until false;
end;

procedure TrRTFParser.ParseText(rtf: string);
var
  temp: string;
  i, p, max, BracketNum: integer;
begin
  TempTags.Clear;
  p:=1;
  max:=Length(rtf);
  ResStr:='';
  BracketNum:=0;
  repeat
    if p > max then
      Break;
    Ch1:=rtf[p];
    if p < max then
      Ch2:=rtf[p + 1]
    else
      Ch2:=' ';

    // actual char - special action for special char
    case Ch1 of
      '{':
        begin
          // begin of block
          AddChar;
          Inc(BracketNum);
        end;
      '}':
        begin
          Dec(BracketNum);
          if BracketNum < 0 then
          begin
            // end of file
            AddResStr;
            Break;
          end;
          AddChar;
          if BracketNum = 0 then
          begin
            // end of last block
            AddResStr;
            IsTagActive:=false;
          end;
        end;
      #13, #10:
        if BracketNum = 0 then
          CheckTagStop;

      '\':
        begin
          if BracketNum > 0 then
            AddChar
          else
            // next char after \
            case Ch2 of
              '\', '{', '}':
                begin
                  if ResStr = '' then
                    ResStr:=ResStr + ' ';
                  ResStr:=ResStr + Ch2;
                  Inc(p);
                end;
              '''':
                begin
                  // hexa num
                  temp:=copy(rtf, p + 2, 2);
                  i:=StrToIntDef('$' + temp, 0);
                  if i > 0 then
                    ResStr:=ResStr + string(AnsiChar(Chr(i)));
                  Inc(p, 3);
                end;
            else
              begin
                // new tag
                AddResStr;
                ResStr:=Ch1;
                IsTagActive:=true;
              end;
            end;
        end;
      ' ':
        begin
          if BracketNum > 0 then
            AddChar
          else
            if not CheckTagStop then
              AddChar;
        end;
    else
      ResStr:=ResStr + Ch1;
    end;
    Inc(p);
  until false;
end;

procedure TrRTFParser.ParseLines(SupportedTags: TStrings = nil);
var
  i: integer;
  t: string;
  L: TStrings;

  function IsSupportedTag: boolean;
  var
    si: integer;
    st: string;
  begin
    Result:=false;
    for si:=0 to SupportedTags.Count - 1 do
    begin
      st:=SupportedTags[si];
      if copy(st, Length(st), 1) = '*' then
        // any char on * place
        Result:=AnsiStartsText(copy(st, 1, Length(st) - 1), t)
      else
        if (copy(st, Length(st), 1) = '#') and AnsiStartsText(copy(st, 1, Length(st) - 1), t) then
        begin
          // any number on # place
          Result:=StrToIntDef(copy(t, Length(st), 255), -1) >= 0;
        end
        else
          Result:=SameText(st, t);
      if Result then
        Exit;
    end;
  end;

begin
  L:=TStringList.Create;
  try
    L.Assign(TempTags);
    // proc all lines and parse again special tags
    for i:=0 to L.Count - 1 do
    begin
      t:=L[i];
      if AnsiStartsText(cRTFRtlch, t) then
      begin
        // parse special tag
        ParseText(copy(t, 2, Length(t)));
        FRtfElements.AddStrings(TempTags);
      end
      else
        FRtfElements.Add(t);
    end;

    // proc all lines, extract unsupported tags and jion texts in more lines without tags
    L.Assign(FRtfElements);
    TempTags.Clear;
    ResStr:='';
    for i:=0 to L.Count - 1 do
    begin
      t:=L[i];
      if t <> '' then
      begin
        if CharInSet(t[1], ['\', '{']) then
        begin
          if AnsiContainsText(t, cRTFFontTbl) then
            ParseFontTbl(t)
          else
            if (SupportedTags = nil) or IsSupportedTag then
            begin
              AddResStr;
              TempTags.Add(t);
            end;
        end
        else
        begin
          ResStr:=ResStr + t;
        end;
      end;
    end;
    AddResStr;

    FRtfElements.Assign(TempTags);
  finally
    L.Free;
  end;
end;

end.
